syntax = "proto3";

package hiku.protobuf.query;

import "google/protobuf/struct.proto";

message Field {
  string name = 1;
  google.protobuf.Struct options = 3;
}

message Link {
  string name = 1;
  Node node = 2;
  google.protobuf.Struct options = 4;
}

message Item {
  oneof value {
    Field field = 1;
    Link link = 2;
  }
}

message Node {
  repeated Item items = 1;
}
